/*
 * PIC12F1822 MCU with I2C LCD and DHT11 sensor.
 * C Code for mikroC PRO for PIC compiler.
 * Internal oscillator used @ 8MHz.
 * This is a free software with NO WARRANTY.
 * https://simple-circuit.com/
*/


// DHT11 pin connection (here data pin is connected to pin RA4)
#define DHT11_PIN         RA4_bit
#define DHT11_PIN_DIR     TRISA4_bit

// include I2C LCD driver source file
#include "I2C_LCD.c"

// send start signal to the sensor
void Start_Signal(void)
{
  DHT11_PIN     = 0;   // connection pin output low
  DHT11_PIN_DIR = 0;   // configure connection pin as output
  delay_ms(25);        // wait 25 ms
  DHT11_PIN     = 1;   // connection pin output high
  delay_us(30);        // wait 30 us
  DHT11_PIN_DIR = 1;   // configure connection pin as input
}

// check sensor response
char Check_Response()
{
  TMR1L = TMR1H = 0;  // reset Timer1 registers

  // wait until DHT11_PIN becomes high (checking of 80s low time response)
  while(!DHT11_PIN && TMR1L < 100) ;

  if(TMR1L >= 100)  // if response time > 80S  ==> response error
    return 0;       // return 0 (device has a problem with response)

  TMR1L = TMR1H = 0;  // reset Timer1 registers

  // wait until DHT11_PIN becomes low (checking of 80s high time response)
  while(DHT11_PIN && TMR1L < 100) ;

  if(TMR1L >= 100)  // if response time > 80S  ==> response error
    return 0;       // return 0 (device has a problem with response)

  return 1;   // return 1 (response OK)
}

// data read function
void Read_Data(unsigned short* dht_data)
{
  char i;
  *dht_data = 0;
  for(i = 0; i < 8; i++)
  {
    TMR1L = TMR1H = 0;  // reset Timer1 registers

    while(!DHT11_PIN)   // wait until DHT11_PIN becomes high
    {
      if(TMR1L > 80)    // if low time > 80S  ==>  Time out error (Normally it takes 50s)
        return;         // return (timeout error)
    }

    TMR1L = TMR1H = 0;  // reset Timer1 registers

    while(DHT11_PIN)    // wait until DHT11_PIN becomes low
    {
      if(TMR1L > 80)  // if high time > 80S  ==>  Time out error (normally it takes 26-28s for 0 and 70s for 1)
        return;       // return (timeout error)
    }
    if(TMR1L > 40)                   // if high time > 40S  ==>  sensor sent 1
      *dht_data |= (1 << (7 - i));   // set bit (7 - i)
  }

  return;   // return (data read OK)
}

char* codetxt_to_ramtxt(const char* ctxt)
{
  static char text_[9];
  char i;
  for(i = 0; text_[i] = ctxt[i]; i++);
  return text_;
}

// main function
void main()
{
  OSCCON = 0x70;  // set internal oscillator to 8MHz
  ANSELA = 0;     // configure all PORTA pins as digital
  T1CON  = 0x10;  // set Timer1 clock source to internal with 1:2 prescaler (Timer1 clock = 1MHz)
  TMR1L  = TMR1H = 0;  // reset Timer1 registers

  I2C1_Init(100000);  // initialize I2C bus with clock frequency of 100kHz
  LCD_Begin(0x4E);    // initialize LCD module with I2C address = 0x4E

  while(1)
  {
    char T_Byte1, T_Byte2, H_Byte1, H_Byte2, CheckSum;
    char temp[] = "Temp = 00.0 C";
    char humi[] = "RH   = 00.0 %";

    TMR1ON_bit = 1;       // enable Timer1 module
    Start_Signal();       // send a start signal to the sensor
    if(Check_Response())  // check if there is a response from sensor (if OK start reading humidity and temperature data)
    {   // response OK ==> read (and save) data from the DHT11 sensor
      Read_Data(&H_Byte1);  // read humidity 1st byte
      Read_Data(&H_Byte2);  // read humidity 2nd byte
      Read_Data(&T_Byte1);   // read temperature 1st byte
      Read_Data(&T_Byte2);   // read temperature 2nd byte
      Read_Data(&CheckSum);  // read checksum
      TMR1ON_bit = 0;        // disable Timer1 module

      // test if all data were sent correctly
      if( CheckSum == ((H_Byte1 + H_Byte2 + T_Byte1 + T_Byte2) & 0xFF) )
      {
        temp[7]  = T_Byte1 / 10  + '0';
        temp[8]  = T_Byte1 % 10  + '0';
        temp[10] = T_Byte2 / 10  + '0';
        humi[7]  = H_Byte1 / 10  + '0';
        humi[8]  = H_Byte1 % 10  + '0';
        humi[10] = H_Byte2 / 10  + '0';
        temp[11] = 223;    // put degree symbol ()
        LCD_Goto(1, 1);    // move cursor to column 1, row 1
        LCD_Print(temp);
        LCD_Goto(1, 2);    // move cursor to column 1, row 2
        LCD_Print(humi);
      }
      else
      {
        LCD_Cmd_(_LCD_CLEAR);  // clear LCD
        LCD_Goto(5, 1);        // move cursor to column 1, row 2
        LCD_Print( codetxt_to_ramtxt("Checksum") );
        LCD_Goto(6, 2);        // move cursor to column 1, row 2
        LCD_Print( codetxt_to_ramtxt("Error") );
      }
    }

    else
    {
      LCD_Cmd_(_LCD_CLEAR);  // clear LCD
      LCD_Goto(6, 1);        // move cursor to column 1, row 2
      LCD_Print( codetxt_to_ramtxt("Sensor") );
      LCD_Goto(6, 2);        // move cursor to column 6, row 2
      LCD_Print( codetxt_to_ramtxt("Error") );
    }

    delay_ms(1000);   // wait a second

  }

}

// end of code.